/*
Copyright (c) 2010, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "barrier.h"
#include "portable.h"
#include "smprc.h"
#include "smpasm.h"
#include "smpequ.h"

#define MAX_LOGICAL_CPU 256
#define MAX_STACK_SIZE 512
#define MWAIT_ALIGN 64

// Memory-mapped APIC Offsets
#define APIC_LOCAL_APIC_ID 0x020
#define APIC_EOI 0xB0
#define APIC_ICR_LO 0x300
#define APIC_ICR_HI 0x310
#define APIC_TMR_LVT 0x320
#define APIC_LINT0_LVT 0x350
#define APIC_TMR_INITIAL_CNT 0x380
#define APIC_TMR_CURRENT_CNT 0x390
#define APIC_TMR_DIVIDE_CFG 0x3E0

#define MSR_APIC_BASE 0x1B
#define MSR_EXT_XAPIC_LOGICAL_APIC_ID 0x802
#define MSR_APIC_EOI 0x80B
#define MSR_EXT_XAPIC_ICR 0x830
#define MSR_APIC_TMR_LVT 0x832
#define MSR_APIC_LINT0_LVT 0x835
#define MSR_APIC_TMR_INITIAL_CNT 0x838
#define MSR_APIC_TMR_CURRENT_CNT 0x839
#define MSR_APIC_TMR_DIVIDE_CFG 0x83E

typedef struct smp_setup_options {
    U32 opcode;
    U32 apicid;
    void *param;
    CALLBACK function;
} SMP_SETUP_OPTIONS;

typedef struct cpu_data {
    U32 stack[MAX_STACK_SIZE];
    U32 use_mwait;
    U32 mwait_hint;
    U32 int_break_event;
    U32 status;
    CALLBACK function;
    void *param;
} CPU_DATA;

typedef struct idtr {
    U16 limit;
    U32 base;
} attr_packed IDTR;

typedef struct exception_info {
    U32 gpf_installed;
    U32 OldGpfAddress;
    U16 OldGpfAddressSel;
    IDTR idt_descriptor;
    U64 idt_table[0x40];
} EXCEPTION_INFO;

#define SMP_MAGIC 0x69534D50

typedef struct smp_host {
    U32 initialized;
    U32 mem_region_below_1M;
    U32 real_mode_callback_addr;
    U32 logical_processor_count;
    U32 expected_processor_count;
    U32 bclk;
    EXCEPTION_INFO bsp_exception_info;
    EXCEPTION_INFO ap_exception_info;
    asmlinkage void (*wait_for_control)(U32 *, U32, U32, U32, U32);
    U8 *control;
    CPU_INFO cpu[MAX_LOGICAL_CPU];
    CPU_DATA cpu_data[MAX_LOGICAL_CPU];
    U8 control_region[MWAIT_ALIGN * MAX_LOGICAL_CPU + MWAIT_ALIGN];
} SMP_HOST;

static void read_apicid(void *param);
static asmlinkage void find_logical_processors(void *param);
static asmlinkage void mp_worker(void *param);
static void prepare_mp_worker(void *param);
static void ap_park(void *param);

static U32 find_processor_id_for_this_cpu(U32 * processor_id, SMP_HOST * host);
static U32 find_processor_id_for_this_apicid(U32 apicid, U32 * processor_id, SMP_HOST * host);

static const IDTR real_mode_idtr = { .limit = 0x3ff, .base = 0 };

static void InitSipiCode(U32 output_address, U32 function, U32 param)
{
    // Move SIPI code below 1M
    memcpy((void *)output_address, pm32, pm32_size);

    *(volatile U32 *)(U32) (output_address + FUNCTIONPTR) = function;
    *(volatile U32 *)(U32) (output_address + PARAM) = param;
    *(volatile U32 *)(U32) (output_address + AWAKE) = 0;
    *(volatile U32 *)(U32) (output_address + BLOCK) = 0;
    *(volatile U32 *)(U32) (output_address + ASLEEP) = 0;
    *(volatile U32 *)(U32) (output_address + BARRIER) = 1;
}

void rdmsr64(U32 msr, U64 * data_addr, U32 * status)
{
    U32 lo_data, hi_data;

    rdmsr32(msr, &lo_data, &hi_data, status);
    *data_addr = ((U64) hi_data << 32) + lo_data;
}

void wrmsr64(U32 msr, U64 data, U32 * status)
{
    wrmsr32(msr, (U32) data, (U32) (data >> 32), status);
}

static U32 x2apic_enabled(void)
{
    U64 temp64;
    U32 status;

    rdmsr64(MSR_APIC_BASE, &temp64, &status);

    return (U32) (temp64 & (1 << 10)) ? 1 : 0;
}

static U32 get_apicbase(void)
{
    U64 temp64;
    U32 status;

    rdmsr64(MSR_APIC_BASE, &temp64, &status);

    return (U32) (temp64 & 0xfffff000);
}

static void send_apicmsg(U32 msgdata, U32 apicid)
{
    U32 ICRHigh = 0;
    U32 ICRLow = 0;

    if (x2apic_enabled()) {
        U32 status;

        ICRHigh = apicid;
        ICRLow = msgdata;
        wrmsr32(MSR_EXT_XAPIC_ICR, ICRLow, ICRHigh, &status);
    } else {
        U32 ApicBase = get_apicbase();

        ICRHigh = apicid << 24;
        ICRLow = msgdata;

        *(volatile U32 *)(U32) (ApicBase + APIC_ICR_HI) = ICRHigh;
        *(volatile U32 *)(U32) (ApicBase + APIC_ICR_LO) = ICRLow;
    }

    return;
}

static U32 get_apic_ICRLow(void)
{
    if (x2apic_enabled()) {
        U64 temp64;
        U32 status;
        rdmsr64(MSR_EXT_XAPIC_ICR, &temp64, &status);
        return (U32) temp64;
    } else {
        return (*(volatile U32 *)(U32) (get_apicbase() + APIC_ICR_LO));
    }
}

static void send_apic_eoi(void)
{
    U32 status;
    if (x2apic_enabled())
        wrmsr32(MSR_APIC_EOI, 0, 0, &status);
    else
        *(volatile U32 *)(U32) (get_apicbase() + APIC_EOI) = 0;
}

static void setup_apic(bool mask, U8 vector)
{
    U32 status;
    U32 lvt = (mask ? 1 << 16 : 0) | vector;
    if (x2apic_enabled())
        wrmsr32(MSR_APIC_TMR_LVT, lvt, 0, &status);
    else
        *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_LVT) = lvt;
}

static bool mask_lint0(bool mask)
{
    bool oldmask;
    U32 val, dummy;

    if (x2apic_enabled())
        rdmsr32(MSR_APIC_LINT0_LVT, &val, &dummy, &dummy);
    else
        val = *(volatile U32 *)(U32) (get_apicbase() + APIC_LINT0_LVT);
    oldmask = (val >> 16) & 1;
    val &= ~(1 << 16);
    val |= ((U32) mask) << 16;
    if (x2apic_enabled())
        wrmsr32(MSR_APIC_LINT0_LVT, val, 0, &dummy);
    else
        *(volatile U32 *)(U32) (get_apicbase() + APIC_LINT0_LVT) = val;
    return oldmask;
}

//-----------------------------------------------------------------------------
static void start_apic_timer(struct smp_host *host, U32 timeout_in_usecs)
{
    // APIC Timer runs at bclk MHz and by default decrements
    // the current count register at once per two clocks.
    // t = time in milliseconds
    // c = APIC Timer Initial Value
    // c = (t * 10^(-6) sec) * (bclk * 10^6 count/sec) * (1/2 clocks)
    // Notice seconds and exponents cancel out leaving count value
    // c = (t * bclk / 2)

    // Prepare to use APIC Timer
    // 1. Get APIC memory base address
    // 2. Start APIC Timer by writing inital count register
    // 3. Set APIC memory base address + APIC_TMR_INITIAL_CNT
    //    as the time in usecs * APIC timer rate / 2

    U32 status;

    if (x2apic_enabled())
        wrmsr32(MSR_APIC_TMR_INITIAL_CNT, (U32) (timeout_in_usecs * host->bclk / 2), 0, &status);
    else
        *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_INITIAL_CNT) = timeout_in_usecs * host->bclk / 2;

    return;
}

//-----------------------------------------------------------------------------
static U32 get_apic_timer(void)
{
    if (x2apic_enabled()) {
        U32 status, temp32, dummy;
        rdmsr32(MSR_APIC_TMR_CURRENT_CNT, &temp32, &dummy, &status);
        return temp32;
    } else
        return (*(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_CURRENT_CNT));
}

static void send_broadcast_init(struct smp_host *host)
{
    U32 temp32;

    // Send the INIT assert
    temp32 = (3ul << 18) | // All excluding self
             (1ul << 14) | // LEVEL is 1
             (5ul << 8);   // INIT

    send_apicmsg(temp32, (U32) 0xff);

    if (!x2apic_enabled()) {
        // start timer 20us
        start_apic_timer(host, 20);

        // while send is still pending and timeout not expired
        while ((get_apic_ICRLow() & (1ul << 12)) && get_apic_timer());
    }

    // Send the INIT de-assert
    temp32 = (3ul << 18) | // All excluding self
             (1ul << 15) | // TRIGGER is 1
             (5ul << 8);   // INIT

    send_apicmsg(temp32, (U32) 0xff);

    if (!x2apic_enabled()) {
        // start timer 20us
        start_apic_timer(host, 20);

        // while send is still pending and timeout not expired
        while ((get_apic_ICRLow() & (1 << 12)) && get_apic_timer());
    }

    // wait for 10ms for INIT processing to complete
    //start timer 10ms
    start_apic_timer(host, 10000);
    while (get_apic_timer());

    return;
}

static void send_broadcast_sipi(struct smp_host *host, U32 addr)
{
    U32 temp32;

    // Send the SIPI
    temp32 = (3ul << 18) | // All excluding self
             (1ul << 14) | // LEVEL is 1
             (6ul << 8);   // SIPI
    addr >>= 12;
    addr &= 0xff;
    temp32 |= addr;
    send_apicmsg(temp32, (U32) 0xff);

    if (!x2apic_enabled()) {
        // start timer 20us
        start_apic_timer(host, 20);

        // While send is still pending or timeout not expired
        while ((get_apic_ICRLow() & (1ul << 12)) && get_apic_timer());
    }

    // While timeout not expired
    // start timer 20us
    start_apic_timer(host, 20);
    while (get_apic_timer());

    // Send it again (MP Spec 1.4 requirement)
    send_apicmsg(temp32, (U32) 0xff);

    if (!x2apic_enabled()) {
        // start timer 20us
        start_apic_timer(host, 20);

        // While send is still pending and timeout not expired
        while ((get_apic_ICRLow() & (1ul << 12)) && get_apic_timer());
    }

    return;
}

static U32 apic_enabled(void)
{
    U64 temp64;
    U32 status;

    rdmsr64(MSR_APIC_BASE, &temp64, &status);

    return (U32) (temp64 & (1 << 11)) ? 1 : 0;
}

//-----------------------------------------------------------------------------
static U32 do_callback(struct smp_host *host, ASM_CALLBACK function, void *param)
{
    U32 addr;
    U32 thread_count = host->expected_processor_count;

    if (!apic_enabled()) {
        dprintf("smp", "APIC is not enabled, returning 1 as status\n");
        return 1;
    }

    addr = host->mem_region_below_1M;

    if (thread_count == 1)
        return 1;

    InitSipiCode(addr, (U32) function, (U32) param);
    send_broadcast_init(host);
    send_broadcast_sipi(host, addr);

    thread_count--; // We already have the BSP

    // We already know how many processors we have, so wait for that many to check in.
    while (*(volatile U32 *)(U32) (addr + ASLEEP) != thread_count)
        pause32();

    *(volatile U32 *)(U32) (addr + BARRIER) = 0;

    return (1);
}

//-----------------------------------------------------------------------------
void read_apicid(void *param)
{
    // Find the APIC ID
    if (x2apic_enabled()) {
        U64 temp64;
        U32 status;

        rdmsr64(MSR_EXT_XAPIC_LOGICAL_APIC_ID, &temp64, &status);
        *(U32 *) param = (U32) temp64;
    } else if (apic_enabled()) {
        *(U32 *) param = (*(volatile U32 *)(U32) (get_apicbase() + APIC_LOCAL_APIC_ID)) >> 24;
    } else {
        U32 dummy;
        U32 ebx;
        cpuid32(1, &dummy, &ebx, &dummy, &dummy);
        *(U32 *) param = ebx >> 24;
    }
}

static void init_bsp_exception_handling(struct smp_host *host)
{
    struct exception_info *e = &host->bsp_exception_info;

    // If the IDT base is zero, then a real mode IDT is still in use
    if (get_idt_base() == 0) {
        U32 i;

        // Create blank 32-bit protected mode IDT and a descriptor that points to it
        for (i = 0; i < sizeof(e->idt_table) / sizeof(e->idt_table[0]); i++)
            e->idt_table[i] = 0;
        e->idt_descriptor.limit = sizeof(e->idt_table) - 1;
        e->idt_descriptor.base = (U32) e->idt_table;

        dprintf("smp", "new IDT base: %08x new IDT Limit: %04x\n", e->idt_descriptor.base, e->idt_descriptor.limit);

        dprintf("smp", "IDT base: %08x IDT Limit: %04x\n", get_idt_base(), get_idt_limit());

        dprintf("smp", "Setup a blank 32-bit PM IDT for BSP\n");
        set_idtr(&e->idt_descriptor);
        set_protected_mode_exception_handler(0xd, (unsigned long)gpfHandler, my_cs());
        set_idtr(&real_mode_idtr);
        e->OldGpfAddress = e->OldGpfAddressSel = 0;
    } else {
        get_idtr(&e->idt_descriptor);
        get_protected_mode_exception_handler(0xd, &e->OldGpfAddress, &e->OldGpfAddressSel);
        set_protected_mode_exception_handler(0xd, (unsigned long)gpfHandler, my_cs());
    }

    e->gpf_installed = 1;
}

static void init_ap_exception_handling(struct smp_host *host)
{
    struct exception_info *e = &host->ap_exception_info;
    U32 i;

    if (!e->gpf_installed) {
        // Create blank 32-bit protected mode IDT and a descriptor that points to it
        for (i = 0; i < sizeof(e->idt_table) / sizeof(e->idt_table[0]); i++)
            e->idt_table[i] = 0;
        e->idt_descriptor.limit = sizeof(e->idt_table) - 1;
        e->idt_descriptor.base = (U32) e->idt_table;

        // Setup a blank 32-bit protected mode IDT for use on the AP
        set_idtr(&e->idt_descriptor);
        set_protected_mode_exception_handler(0xd, (unsigned long)gpfHandler, my_cs());
        e->gpf_installed = 1;
    } else
        set_idtr(&e->idt_descriptor);
}

//-----------------------------------------------------------------------------
asmlinkage void find_logical_processors(void *param)
{
    SMP_HOST *host = param;
    U32 processor_id = 0;

    init_ap_exception_handling(host);

    if (host->logical_processor_count == host->expected_processor_count)
        while (1) {
        }

    processor_id = host->logical_processor_count++;
    read_apicid(&host->cpu[processor_id].apicid);

    host->cpu[processor_id].present = 1;

    prepare_mp_worker(param);
}

static bool mwait_supported(void)
{
    U32 eax, ecx, dummy;
    cpuid32(0, &eax, &dummy, &dummy, &dummy);
    if (eax < 5)
        return false;
    cpuid32(1, &dummy, &dummy, &ecx, &dummy);
    if (!(ecx & (1 << 3)))
        return false;
    return true;
}

static bool int_break_event_supported(void)
{
    U32 ecx, dummy;
    if (!mwait_supported())
        return false;
    cpuid32(5, &dummy, &dummy, &ecx, &dummy);
    return ecx & (1 << 1) ? true : false;
}

void smp_set_mwait_with_memory(void *working_memory, U32 apicid, bool use_mwait, U32 mwait_hint, U32 int_break_event)
{
    U32 processor_id;
    CPU_DATA *cpu_data;

    struct smp_host *host = working_memory;
    if (!host || host->initialized != SMP_MAGIC)
        return;

    if (find_processor_id_for_this_apicid(apicid, &processor_id, host) == 0)
        return;

    cpu_data = &host->cpu_data[processor_id];

    cpu_data->use_mwait = use_mwait;
    cpu_data->mwait_hint = mwait_hint;
    cpu_data->int_break_event = int_break_event;
}

//-----------------------------------------------------------------------------
asmlinkage void mp_worker(void *param)
{
    SMP_HOST *host = param;
    U32 processor_id = 0;
    U32 *my_control;
    CPU_DATA *cpu_data;

    drop_ap_lock(host->mem_region_below_1M);

    if (find_processor_id_for_this_cpu(&processor_id, host) == 0)
        do_nothing();

    cpu_data = &host->cpu_data[processor_id];
    my_control = (U32 *) (host->control + processor_id * MWAIT_ALIGN);

    for (;;) {
        /* Detect the ability to use mwait every time, just in case the function does something to disable it. */
        host->wait_for_control(my_control, AP_IN_CONTROL,
                               cpu_data->use_mwait && mwait_supported(), cpu_data->mwait_hint,
                               cpu_data->int_break_event && int_break_event_supported());

        // Do assigned function
        cpu_data->function(cpu_data->param);

        // Save results, modify flags, etc is done by the function

        set_control(my_control, BSP_IN_CONTROL);
    }
}

//-----------------------------------------------------------------------------
void prepare_mp_worker(void *param)
{
    SMP_HOST *host = param;
    U32 processor_id = 0;
    CPU_DATA *cpu_data;
    U32 stack_top;

    if (find_processor_id_for_this_cpu(&processor_id, host) == 0)
        do_nothing();

    cpu_data = &host->cpu_data[processor_id];

    cpu_data->status = 1;

    // Switch stacks
    stack_top = (U32) (&host->cpu_data[processor_id].stack[MAX_STACK_SIZE]);

    switch_stack_and_call((U32) mp_worker, param, stack_top);
}

//-----------------------------------------------------------------------------
U32 find_processor_id_for_this_apicid(U32 apicid, U32 * processor_id, SMP_HOST * host)
{
    U32 i;

    for (i = 0; i < host->logical_processor_count; i++)
        if (host->cpu[i].apicid == apicid) {
            *processor_id = i;
            return 1;
        }

    return 0;
}

//-----------------------------------------------------------------------------
U32 find_processor_id_for_this_cpu(U32 * processor_id, SMP_HOST * host)
{
    U32 apicid;

    read_apicid(&apicid);

    return find_processor_id_for_this_apicid(apicid, processor_id, host);
}

//-----------------------------------------------------------------------------
void ap_park(void *param)
{
    (void)param;
}

/* Minimal ACPI parsing code to find and process the MADT */

static U8 GetChecksum(void *mem_addr, U32 mem_size)
{
    U8 *current = mem_addr;
    U8 *end = current + mem_size;
    U8 checksum = 0;

    for (; current < end; current++)
        checksum = checksum + *current;

    return (checksum);
}

#define NAMESEG(s)   (((U32)(s[0]) << 0)  \
                     |((U32)(s[1]) << 8)  \
                     |((U32)(s[2]) << 16) \
                     |((U32)(s[3]) << 24))

#define NAMESEG64(s) (((U64)(s[0]) << 0)   \
                     |((U64)(s[1]) << 8)   \
                     |((U64)(s[2]) << 16)  \
                     |((U64)(s[3]) << 24)  \
                     |((U64)(s[4]) << 32)  \
                     |((U64)(s[5]) << 40)  \
                     |((U64)(s[6]) << 48)  \
                     |((U64)(s[7]) << 56)) \

struct acpi_table_rsdp {
    U8 Signature[8]; // ACPI signature, contains "RSD PTR "
    U8 Checksum; // ACPI 1.0 checksum
    U8 OemId[6]; // OEM identification
    U8 Revision; // Must be (0) for ACPI 1.0 or (2) for ACPI 2.0+
    U32 RsdtPhysicalAddress; // 32-bit physical address of the RSDT
    U32 Length; // Table length in bytes, including header (ACPI 2.0+)
    U64 XsdtPhysicalAddress; // 64-bit physical address of the XSDT (ACPI 2.0+)
    U8 ExtendedChecksum; // Checksum of entire table (ACPI 2.0+)
    U8 Reserved[3]; // Reserved, must be zero
};

#define ACPI_RSDP_REV0_SIZE 20 // Size of original ACPI 1.0 RSDP

struct acpi_table_header {
    U8 Signature[4]; // ASCII table signature
    U32 Length; // Length of table in bytes, including this header
    U8 Revision; // ACPI Specification minor version #
    U8 Checksum; // To make checksum of entire table == 0
    U8 OemId[6]; // ASCII OEM identification
    U8 OemTableId[8]; // ASCII OEM table identification
    U32 OemRevision; // OEM revision number
    U8 AslCompilerId[4]; // ASCII ASL compiler vendor ID
    U32 AslCompilerRevision; // ASL compiler version
};

struct acpi_table_rsdt {
    struct acpi_table_header Header; // Common ACPI table header
    U32 TableOffsetEntry[1]; // Array of pointers to ACPI tables
};

struct acpi_subtable_header {
    U8 Type;
    U8 Length;
};

struct acpi_table_madt {
    struct acpi_table_header Header; // Common ACPI table header
    U32 Address; // Physical address of local APIC
    U32 Flags;
};

enum AcpiMadtType {
    ACPI_MADT_TYPE_LOCAL_APIC = 0,
    ACPI_MADT_TYPE_X2APIC = 9,
};

#define ACPI_MADT_ENABLED (1) // 00: Processor is usable if set

struct acpi_madt_local_apic {
    struct acpi_subtable_header Header;
    U8 ProcessorId; // ACPI processor id
    U8 Id; // Processor's local APIC id
    U32 LapicFlags;
};

struct acpi_madt_x2apic {
    struct acpi_subtable_header Header;
    U16 Reserved; // Must be zero
    U32 x2apicId; // Processor's X2APIC ID
    U32 x2apicFlags;
    U32 UID;
};

static void *GetRsdtPointer(void *mem_addr, U32 mem_size)
{
    U8 *current = mem_addr;
    U8 *end = current + mem_size;

    // Quick sanity check for a valid start address
    if (!current)
        return NULL;

    for (; current < end; current += 16)
        if (*(U64 *) current == NAMESEG64("RSD PTR "))
            if (GetChecksum(current, ACPI_RSDP_REV0_SIZE) == 0) {
                // RSD pointer structure checksum okay, lookup the RSDT pointer.
                struct acpi_table_rsdp *rsdp = (struct acpi_table_rsdp *)current;
                return (void *)rsdp->RsdtPhysicalAddress;
            }

    return NULL;
}

static U32 get_num_tables(struct acpi_table_rsdt *rsdt)
{
    // Compute number of table pointers included in RSDT
    return (rsdt->Header.Length - sizeof(struct acpi_table_header)) / sizeof(U32);
}

static void *GetTablePtr(struct acpi_table_rsdt *rsdt, U32 signature)
{
    U32 index;
    U32 num_tables;
    struct acpi_table_header **table_array = (struct acpi_table_header **)rsdt->TableOffsetEntry;

    // Compute number of table pointers included in RSDT
    num_tables = get_num_tables(rsdt);

    for (index = 0; index < num_tables; index++)
        if ((*(U32 *) (table_array[index]->Signature) == signature) && (GetChecksum(table_array[index], table_array[index]->Length) == 0))
            return table_array[index];

    return NULL;
}

static U32 process_madt(struct acpi_table_madt *madt)
{
    U32 count = 0;
    void *current;
    void *end;

    // Search MADT for Sub-tables with needed data
    current = madt + 1;
    end = (U8 *) madt + madt->Header.Length;

    while (current < end) {
        struct acpi_subtable_header *subtable = current;

        switch (subtable->Type) {
        case ACPI_MADT_TYPE_LOCAL_APIC:
            {
                // Process sub-tables with Type as 0: Processor Local APIC
                struct acpi_madt_local_apic *lapic = current;
                if (lapic->LapicFlags & ACPI_MADT_ENABLED)
                    count++;
                break;
            }
        case ACPI_MADT_TYPE_X2APIC:
            {
                // Process sub-tables with Type as 9: Processor X2APIC
                struct acpi_madt_x2apic *x2apic = current;
                if (x2apic->x2apicFlags & ACPI_MADT_ENABLED)
                    count++;
                break;
            }
        } // switch

        current = (U8 *) subtable + subtable->Length;
    } // while

    return count;
}

static U32 madt_processor_count(void)
{
    struct acpi_table_rsdt *rsdt;
    struct acpi_table_madt *madt;

    // Init memory address as EBDA and scan 1KB region
    rsdt = GetRsdtPointer((void *)(((U32) * (U16 *) 0x40E) << 4), 0x400);

    // Init memory address as E000 segment and scan 64KB region
    if (!rsdt)
        rsdt = GetRsdtPointer((void *)0x0E0000, 0x10000);

    // Init memory address as F000 segment and scan 64KB region
    if (!rsdt)
        rsdt = GetRsdtPointer((void *)0x0F0000, 0x10000);

    if (!rsdt)
        return 0;

    // Find the MADT table which is one of the table pointers in the RSDT
    madt = GetTablePtr(rsdt, NAMESEG("APIC"));
    if (!madt)
        return 0;

    return process_madt(madt);
}

static U32 compute_bclk(void)
{
    U32 status, dummy;
    U32 start, stop;
    U8 temp8;
    U16 delay_count;
    U32 bclk;

#define DELAY_IN_US 1000

    // Compute fixed delay as time
    // delay count = desired time * PIT frequency
    // PIT frequency = 1.193182 MHz
    delay_count = 1193182 / DELAY_IN_US;

    // PIT channel 2 gate is controlled by IO port 0x61, bit 0
#define PIT_CH2_LATCH_REG 0x61
#define CH2_SPEAKER (1 << 1) // bit 1 -- 1 = speaker enabled 0 = speaker disabled
#define CH2_GATE_IN (1 << 0) // bit 0 -- 1 = gate enabled, 0 = gate disabled
#define CH2_GATE_OUT (1 << 5) // bit 5 -- 1 = gate latched, 0 = gate not latched

    // PIT Command register
#define PIT_MODE_COMMAND_REG 0x43
#define SELECT_CH2 (2 << 6)
#define ACCESS_MODE_LOBYTE_HIBYTE (3 << 4)
#define MODE0_INTERRUPT_ON_TERMINAL_COUNT 0 // Despite name, no interrupts on CH2

    // PIT Channel 2 data port
#define PIT_CH2_DATA 0x42

    // Disable the PIT channel 2 speaker and gate
    temp8 = input_u8(PIT_CH2_LATCH_REG);
    temp8 &= ~(CH2_SPEAKER | CH2_GATE_IN);
    output_u8(PIT_CH2_LATCH_REG, temp8);

    // Setup command and mode
    output_u8(PIT_MODE_COMMAND_REG, SELECT_CH2 | ACCESS_MODE_LOBYTE_HIBYTE | MODE0_INTERRUPT_ON_TERMINAL_COUNT);

    // Set time for fixed delay
    output_u8(PIT_CH2_DATA, (U8) (delay_count));
    output_u8(PIT_CH2_DATA, (U8) (delay_count >> 8));

    // Prepare to enable channel 2 gate but leave the speaker disabled
    temp8 = input_u8(PIT_CH2_LATCH_REG);
    temp8 &= ~CH2_SPEAKER;
    temp8 |= CH2_GATE_IN;

    if (x2apic_enabled()) {
        // Set APIC Timer Divide Value as 2
        wrmsr32(MSR_APIC_TMR_DIVIDE_CFG, 0, 0, &status);

        // start APIC timer with a known value
        start = ~0UL;
        wrmsr32(MSR_APIC_TMR_INITIAL_CNT, start, 0, &status);
    }
    else {
        // Set APIC Timer Divide Value as 2
        *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_DIVIDE_CFG) = 0UL;

        // start APIC timer with a known value
        start = ~0UL;
        *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_INITIAL_CNT) = start;
    }

    // Actually start the PIT channel 2
    output_u8(PIT_CH2_LATCH_REG, temp8);

    // Wait for the fixed delay
    while (!(input_u8(PIT_CH2_LATCH_REG) & CH2_GATE_OUT));

    if (x2apic_enabled()) {
        // read the APIC timer to determine the change that occurred over this fixed delay
        rdmsr32(MSR_APIC_TMR_CURRENT_CNT, &stop, &dummy, &status);

        // stop APIC timer
        wrmsr32(MSR_APIC_TMR_INITIAL_CNT, 0, 0, &status);
    }
    else {
        // read the APIC timer to determine the change that occurred over this fixed delay
        stop = *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_CURRENT_CNT);

        // stop APIC timer
        *(volatile U32 *)(U32) (get_apicbase() + APIC_TMR_INITIAL_CNT) = 0UL;
    }

    // Disable channel 2 speaker and gate input
    temp8 = input_u8(PIT_CH2_LATCH_REG);
    temp8 &= ~(CH2_SPEAKER | CH2_GATE_IN);
    output_u8(PIT_CH2_LATCH_REG, temp8);

    bclk = (start - stop) * 2 / DELAY_IN_US;

    // Round bclk to the nearest 100/12 integer value
    bclk = ((((bclk * 24) + 100) / 200) * 200) / 24;
    dprintf("smp", "Compute bclk: %uMHz\n", bclk);
    return bclk;
}

U32 smp_init_with_memory(void *working_memory, void *page_below_1M, void *reserved_mwait_memory)
{
    struct smp_host *host = working_memory;

    /* Sanity checks on the amounts of memory our public interface claims we
       can work within. */
    if (sizeof(struct smp_host) > SMP_WORKING_MEMORY_SIZE) {
        dprintf("smp", "Internal error: SMP_WORKING_MEMORY_SIZE too small; need %u\n", sizeof(struct smp_host));
        return 0;
    }
    if (SMP_RESERVED_MEMORY_SIZE < MAX_LOGICAL_CPU * MWAIT_ALIGN + wait_for_control_asm_size) {
        dprintf("smp", "Internal error: SMP_RESERVED_MEMORY_SIZE too small; need %u\n", MAX_LOGICAL_CPU * MWAIT_ALIGN + wait_for_control_asm_size);
        return 0;
    }

    if (host->initialized == SMP_MAGIC)
        return host->logical_processor_count;

    host->expected_processor_count = madt_processor_count();
    if (host->expected_processor_count == 0)
        host->expected_processor_count = 1;

    dprintf("smp", "Processor count from MADT: %u\n", host->expected_processor_count);

    if (host->expected_processor_count > MAX_LOGICAL_CPU)
        return 0;

    host->mem_region_below_1M = (U32) page_below_1M;
    host->logical_processor_count = 1;
    if (reserved_mwait_memory) {
        host->control = reserved_mwait_memory;
        host->wait_for_control = (void *)(((U8 *) reserved_mwait_memory) + MWAIT_ALIGN * MAX_LOGICAL_CPU);
        memcpy(host->wait_for_control, wait_for_control_asm, wait_for_control_asm_size);
    } else {
        host->control = (U8 *) (((U32) host->control_region - 1) & (MWAIT_ALIGN - 1)) + 1;
        host->wait_for_control = wait_for_control_asm;
    }

    // Init the host structure
    {
        U32 i;
        for (i = 0; i < host->expected_processor_count; i++) {
            host->cpu[i].present = 0;
            set_control((U32 *) (host->control + i * MWAIT_ALIGN), BSP_IN_CONTROL);
            host->cpu_data[i].use_mwait = true;
            host->cpu_data[i].mwait_hint = 0;
            host->cpu_data[i].int_break_event = 1;
            host->cpu_data[i].function = ap_park;
            host->cpu_data[i].param = NULL;
        }
    }

    host->cpu[0].present = 1;
    read_apicid(&host->cpu[0].apicid);

    host->bclk = compute_bclk();

    host->bsp_exception_info.gpf_installed = 0;
    host->ap_exception_info.gpf_installed = 0;
    init_bsp_exception_handling(host);

    if (do_callback(host, find_logical_processors, host)) {
        host->initialized = SMP_MAGIC;
        return host->logical_processor_count;
    } else
        return 0;
}

void smp_phantom_init_with_memory(void *working_memory)
{
    struct smp_host *host = working_memory;
    if (!host || host->initialized != SMP_MAGIC)
        return;
    host->initialized = 0;
}

U32 smp_read_bclk_with_memory(void *working_memory)
{
    struct smp_host *host = working_memory;
    if (!host || host->initialized != SMP_MAGIC)
        return 0;
    return host->bclk;
}

const CPU_INFO *smp_read_cpu_list_with_memory(void *working_memory)
{
    struct smp_host *host = working_memory;
    if (!host || host->initialized != SMP_MAGIC)
        return NULL;
    return host->cpu;
}

U32 smp_function_with_memory(void *working_memory, U32 apicid, CALLBACK function, void *param)
{
    struct smp_host *host = working_memory;
    if (!host || host->initialized != SMP_MAGIC) {
        dprintf("smp", "smp_function returning 0 because working memory not initialized\n");
        return 0;
    }

    if (!function) {
        dprintf("smp", "smp_function returning 0 because !function\n");
        return 0;
    }

    if (apicid == host->cpu[0].apicid) {
        set_idtr(&host->bsp_exception_info.idt_descriptor);
        function(param);
        set_idtr(&real_mode_idtr);
    } else {
        U32 processor_id;
        CPU_DATA *cpu_data;
        U32 *my_control;

        if (find_processor_id_for_this_apicid(apicid, &processor_id, host) == 0) {
            dprintf("smp", "smp_function returning 0 because APIC ID not found\n");
            return 0;
        }

        cpu_data = &host->cpu_data[processor_id];
        my_control = (U32 *) (host->control + processor_id * MWAIT_ALIGN);

        // Check if AP is available - FIXME: this should be an assert
        if (*my_control != BSP_IN_CONTROL) {
            dprintf("smp", "smp_function returning 0 because BSP not in control\n");
            return 0;
        }
        // Assign the function and its parameter
        cpu_data->function = function;
        cpu_data->param = param;

        set_control(my_control, AP_IN_CONTROL);
        host->wait_for_control(my_control, BSP_IN_CONTROL, cpu_data[0].use_mwait && mwait_supported(), cpu_data[0].mwait_hint, cpu_data[0].int_break_event && int_break_event_supported());
    }

    return 1;
}

/* Called from smpasm directly, which won't use a C prototype, so just give one here to silence the warning. */
asmlinkage void intHandler(void);
asmlinkage void intHandler(void)
{
    void (*func) (void *);
    void *param;
    U16 dummy;

    send_apic_eoi();
    get_protected_mode_exception_handler(0x31, (U32 *) &func, &dummy);
    get_protected_mode_exception_handler(0x32, (U32 *) &param, &dummy);
    func(param);
}

static void set_control_callback(void *param)
{
    set_control(param, 0);
}

void smp_sleep_with_memory(void *working_memory, U32 microseconds)
{
    struct smp_host *host = working_memory;
    U32 processor_id;
    CPU_DATA *cpu_data;
    U32 my_control[MWAIT_ALIGN / sizeof(U32)];
    bool oldmask;

    if (!host || host->initialized != SMP_MAGIC)
        return;

    if (find_processor_id_for_this_cpu(&processor_id, host) == 0)
        return;

    cpu_data = &host->cpu_data[processor_id];

    set_protected_mode_exception_handler(0x30, (unsigned long)intHandler_asm, my_cs());
    set_protected_mode_exception_handler(0x31, (unsigned long)set_control_callback, 0);
    set_protected_mode_exception_handler(0x32, (unsigned long)my_control, 0);

    set_control(my_control, 1);

    /* Set up the timer that puts the BSP back in control */
    oldmask = mask_lint0(1); /* Need to prevent other interrupts while sti'd */
    enable_interrupts();

    setup_apic(0, 0x30);
    start_apic_timer(host, microseconds);

    host->wait_for_control(my_control, 0,
                           cpu_data->use_mwait && mwait_supported(), cpu_data->mwait_hint,
                           cpu_data->int_break_event && int_break_event_supported());

    setup_apic(1, 0xff);

    disable_interrupts();
    mask_lint0(oldmask);
}
